/*
 * alsa_seqmidi.c - jack midi <-> alsa seq interface
 *
 * Copyright (c) 2006 Dmitry S. Baikov <c0ff@konstruktiv.org>
 *
 * Uses ideas from jack2aseq (C) 2006 Lars Luthman, which in turn
 *   uses ideas and modified code snippets from:
 *                 alsaseq2jackmidi, (C) 2005 Sean Bolton.
 *                 QJackCtl, (C) 2003-2006, rncbc aka Rui Nuno Capela
 * all of which are released under the GNU General Public License.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 */

#include <alsa/asoundlib.h>
#include <jack/jack.h>
#include <jack/midiport.h>
#include <jack/ringbuffer.h>
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <semaphore.h>

#include "./jack_utils.h"
#include "./alsa_midi.h"

typedef struct port_t port_t;

enum {
	PORT_HASH_BITS = 4,
	PORT_HASH_SIZE = 1 << PORT_HASH_BITS
};

typedef struct alsa_seqmidi {
	alsa_midi_t alsa_midi;
	snd_seq_t *alsa_client;
	jack_client_t *jack_client;
	sem_t alsa_sem;
	int alsa_announce;
	int alsa_input_que;
	int alsa_output_que;
	snd_midi_event_t *alsa_decoder;
	jack_ringbuffer_t *midi_in;
	jack_ringbuffer_t *midi_out;
	jack_ringbuffer_t *dead4alsa;
	jack_ringbuffer_t *dead4jack;
	unsigned char alsa_event_buf[1024];
	unsigned char jack_event_buf[1024];
	port_t *ports[2][PORT_HASH_SIZE];
} alsa_seqmidi_t;


alsa_midi_t* alsa_seqmidi_new(const char *client_name, int hw_only);


typedef struct _port_type {
	int alsa_mask;
	int alsa_caps;
	int jack_caps;
	char name[16];
} port_type_t;

struct port_t {
	port_t *next;
	snd_seq_addr_t remote;
	int alsa_port;
	jack_port_t *jack_port;
	void *jack_buf;
	snd_midi_event_t *alsa_encoder;
};

enum { PORT_INPUT = 0, PORT_OUTPUT = 1 };

static port_type_t port_type[2] = {
	{
		SND_SEQ_PORT_CAP_SUBS_READ,
		SND_SEQ_PORT_CAP_WRITE|SND_SEQ_PORT_CAP_SUBS_WRITE/*|SND_SEQ_PORT_CAP_NO_EXPORT*/,
		JackPortIsOutput|JackPortIsPhysical,
		"midi_in"
	},
	{
		SND_SEQ_PORT_CAP_SUBS_WRITE,
		SND_SEQ_PORT_CAP_READ|SND_SEQ_PORT_CAP_SUBS_READ/*|SND_SEQ_PORT_CAP_NO_EXPORT*/,
		JackPortIsInput|JackPortIsPhysical|JackPortIsTerminal,
		"midi_out"
	}
};

typedef struct {
	port_t *port;
	jack_nframes_t time;
	size_t size;
} midi_event_t;

static inline int port_hash(int client, int port)
{
	return (client + port) % PORT_HASH_SIZE;
}

static port_t** port_find(alsa_seqmidi_t *self, int type, int client_id, int port_id)
{
	port_t **pport = &self->ports[type][port_hash(client_id, port_id)];
	while (*pport) {
		port_t *port = *pport;
		if ( port->remote.client == client_id && port->remote.port == port_id)
			return pport;
		pport = &port->next;
	}
	return 0;
}

static void port_insert(alsa_seqmidi_t *self, int t, port_t *port)
{
	port_t **pport = &self->ports[t][port_hash(port->remote.client, port->remote.port)];
	port->next = *pport;
	*pport = port;
}

static void port_remove(port_t **pport)
{
	*pport = (*pport)->next;
}

static void alsa_connect(snd_seq_t *seq, int que, int c1, int p1, int c2, int p2)
{
	snd_seq_port_subscribe_t* sub;
	snd_seq_addr_t seq_addr;
	int err;

	snd_seq_port_subscribe_alloca(&sub);
	seq_addr.client = c1;
	seq_addr.port = p1;
	snd_seq_port_subscribe_set_sender(sub, &seq_addr);
	seq_addr.client = c2;
	seq_addr.port = p2;
	snd_seq_port_subscribe_set_dest(sub, &seq_addr);
//	snd_seq_port_subscribe_set_queue(sub, que);
	if ((err=snd_seq_subscribe_port(seq, sub)))
		printf("can't subscribe %d:%d to %d:%d - %s\n", c1, p1, c2, p2, snd_strerror(err));
}

static port_t *port_create(alsa_seqmidi_t *self, int type, snd_seq_port_info_t *port_info)
{
	char name[64];
	int client_id = snd_seq_port_info_get_client(port_info);
	int port_id = snd_seq_port_info_get_port(port_info);
	const char *port_name = snd_seq_port_info_get_name(port_info);
	port_t *port = malloc (sizeof(port_t));
	port->next = NULL;
	port->remote.client = client_id;
	port->remote.port = port_id;
	port->jack_buf = 0;
	snprintf (name, sizeof(name), "%s (%d:%d) %s", port_type[type].name, client_id, port_id, port_name);
	port->alsa_port = snd_seq_create_simple_port (self->alsa_client, name, port_type[type].alsa_caps, SND_SEQ_PORT_TYPE_APPLICATION);
	port->jack_port = jack_port_register (self->jack_client, name, JACK_DEFAULT_MIDI_TYPE, port_type[type].jack_caps, 0);
	if (type == PORT_INPUT) {
		port->alsa_encoder = 0;
		alsa_connect (self->alsa_client, self->alsa_input_que,
			port->remote.client, port->remote.port, snd_seq_client_id(self->alsa_client), port->alsa_port);
	} else {
		snd_midi_event_new(1024, &port->alsa_encoder);
		snd_midi_event_init(port->alsa_encoder);
		alsa_connect (self->alsa_client, self->alsa_output_que,
			snd_seq_client_id(self->alsa_client), port->alsa_port, port->remote.client, port->remote.port);
	}
	printf("port_create %s\n", name);
	return port;
}

static void port_free(alsa_seqmidi_t *self, port_t *port)
{
	jack_port_unregister(self->jack_client, port->jack_port);
	if (port->alsa_encoder)
		snd_midi_event_free(port->alsa_encoder);
	snd_seq_delete_simple_port (self->alsa_client, port->alsa_port);
	printf("port_free %d:%d\n", port->remote.client, port->remote.port);
	free (port);
}

static void alsa_port_remove(alsa_seqmidi_t *self, int type, port_t **pport)
{
	port_t *dead = *pport;
	port_remove(pport);
	if (jack_ringbuffer_write_space(self->dead4alsa) < sizeof(port_t*)) {
		// store port somewhere
		printf("not enough space in dead4alsa ringbuffer");
	} else
		jack_ringbuffer_write(self->dead4alsa, (char*)&dead, sizeof(dead));
	printf("port_dead %s %d:%d\n", port_type[type].name, dead->remote.client, dead->remote.port);
}

static void alsa_ports_cleanup(alsa_seqmidi_t *self)
{
	port_t *port;
	while (jack_ringbuffer_read_space(self->dead4jack) >= sizeof(port)) {
		jack_ringbuffer_read(self->dead4jack, (char*)&port, sizeof(port));
		port_free(self, port);
	}
}

void alsa_port_update(alsa_seqmidi_t *self, snd_seq_port_info_t *port_info)
{
	int t;
	int client_id = snd_seq_port_info_get_client(port_info);
	int port_id = snd_seq_port_info_get_port(port_info);
	unsigned int port_caps = snd_seq_port_info_get_capability(port_info);
	printf("alsa_port_update %d:%d\n", client_id, port_id);
	if (port_caps & SND_SEQ_PORT_CAP_NO_EXPORT)
		return;
	for(t=0; t<2; ++t) {
		port_t **pport;
		if ((pport = port_find(self, t, client_id, port_id)) && (port_caps & port_type[t].alsa_mask)!=port_type[t].alsa_mask) {
			alsa_port_remove(self, t, pport);
		}
		if (!pport && (port_caps & port_type[t].alsa_mask)==port_type[t].alsa_mask) {
			port_t *port = port_create(self, t, port_info);
			port_insert(self, t, port);
		}
	}
}

void alsa_get_ports(alsa_seqmidi_t *self)
{
	snd_seq_client_info_t *client_info;
	snd_seq_port_info_t *port_info;

	snd_seq_client_info_alloca(&client_info);
	snd_seq_port_info_alloca(&port_info);
	snd_seq_client_info_set_client(client_info, -1);
	while (snd_seq_query_next_client(self->alsa_client, client_info) >= 0)
	{
		int client_id = snd_seq_client_info_get_client(client_info);
		if( client_id == 0 || client_id == snd_seq_client_id(self->alsa_client) )
			continue;
		snd_seq_port_info_set_client(port_info, client_id);
		snd_seq_port_info_set_port(port_info, -1);
		while (snd_seq_query_next_port(self->alsa_client, port_info) >= 0)
		{
			alsa_port_update(self, port_info);
		}
	}
}

void alsa_port_event(alsa_seqmidi_t *self, snd_seq_event_t *ev)
{
	int err;
	int client_id = ev->data.addr.client;
	int port_id = ev->data.addr.port;
	snd_seq_port_info_t *port_info;
	if (ev->data.addr.client == 0 || ev->data.addr.client == snd_seq_client_id(self->alsa_client))
		return;
	switch (ev->type) {
	case SND_SEQ_EVENT_PORT_START:
	case SND_SEQ_EVENT_PORT_CHANGE:
		snd_seq_port_info_alloca(&port_info);
		if ((err=snd_seq_get_any_port_info(self->alsa_client, client_id, port_id, port_info))>=0) {
			alsa_port_update(self, port_info);
			break;
		}
		else printf("can't get port info %d:%d - %s\n", client_id, port_id, snd_strerror(err));
		// else remove port
		// INTENDED FALLTHROUGH
	case SND_SEQ_EVENT_PORT_EXIT: {
		int i;
		for(i=0; i<2; ++i) {
			port_t **pport;
			if ((pport = port_find(self, i, client_id, port_id)))
				alsa_port_remove(self, i, pport);
		}
		break;
		}
	}
}

void alsa_input(alsa_seqmidi_t *self)
{
	int err;
	snd_seq_event_t *ev;
	do {
		long size;
		midi_event_t event;
		port_t **pport;
		jack_ringbuffer_data_t vec[2];

		if ((err = snd_seq_event_input(self->alsa_client, &ev))<0)
			break;
		if (ev->source.client == SND_SEQ_CLIENT_SYSTEM) {
			alsa_port_event(self, ev);
			continue;
		}
		pport = port_find(self, PORT_INPUT, ev->source.client, ev->source.port);
		if (!pport)
			continue;
		event.port = *pport;
		snd_midi_event_init(self->alsa_decoder);
		if ((size = snd_midi_event_decode(self->alsa_decoder, self->alsa_event_buf, sizeof(self->alsa_event_buf), ev))<0)
			continue;
		event.size = size;
		jack_ringbuffer_get_write_vector(self->midi_in, vec);
		if (jack_ringbuffer_vector_space(vec) < (sizeof(event) + event.size))
			continue; // ENOSPC
		event.time = jack_frame_time(self->jack_client) + jack_get_buffer_size(self->jack_client);
		jack_ringbuffer_vector_write(vec, &event, sizeof(event));
		jack_ringbuffer_vector_write(vec, self->alsa_event_buf, event.size);
		jack_ringbuffer_write_advance(self->midi_in, sizeof(event) + event.size);
		printf("midi_in - %d at %d\n", ev->type, (int)event.time);
	} while (err > 0);
}

void alsa_output(alsa_seqmidi_t *self)
{
	snd_seq_event_t alsa_event;
	jack_nframes_t rate = jack_get_sample_rate(self->jack_client);
	snd_seq_real_time_t timestamp;
	while (1) {
		jack_ringbuffer_data_t vec[2];
		midi_event_t event;
		long offset;
		int i, err;

		timestamp.tv_sec = 0;
		timestamp.tv_nsec = 0;

		jack_ringbuffer_get_read_vector(self->midi_out, vec);
		if ( jack_ringbuffer_vector_space(vec) < sizeof(event))
			break; // ENODATA
		jack_ringbuffer_vector_read(vec, &event, sizeof(event));
		if (jack_ringbuffer_vector_space(vec) < event.size)
			break; // ENODATA
		if (event.size > sizeof(self->alsa_event_buf)) {
			jack_ringbuffer_read_advance(self->midi_out, sizeof(event) + event.size);
			continue; // ENOSPC
		}
		jack_ringbuffer_vector_read(vec, self->alsa_event_buf, event.size);
		jack_ringbuffer_read_advance(self->midi_out, sizeof(event) + event.size);
		snd_seq_ev_clear(&alsa_event);
		//snd_midi_event_init(event.port->alsa_encoder);
		if (!snd_midi_event_encode(event.port->alsa_encoder, self->alsa_event_buf, event.size, &alsa_event))
			continue; // EBADF?
		for (i=0; i<event.size; ++i)
			printf("%02x ", (int)self->alsa_event_buf[i]);
		snd_seq_ev_set_source(&alsa_event, event.port->alsa_port);
		snd_seq_ev_set_subs(&alsa_event);
		snd_seq_ev_schedule_real(&alsa_event, self->alsa_output_que, 1, &timestamp);
		offset = (long)((event.time - jack_frame_time(self->jack_client)) * (1000000000.0 / rate));
		printf("offset %ld, event.time = %d", offset, (int)event.time);
		timestamp.tv_sec = offset / 1000000000;
		timestamp.tv_nsec = (offset % 1000000000);
		if ((err=snd_seq_event_output(self->alsa_client, &alsa_event))<0)
			printf("midiout error - %s\n", snd_strerror(err));
		else
			printf("midiout - %d\n", alsa_event.type);
	}
}

static void jack_process_input(alsa_seqmidi_t *self, jack_nframes_t nframes)
{
	jack_nframes_t jack_time = jack_last_frame_time(self->jack_client);
	int i;
	for(i=0; i<PORT_HASH_SIZE; ++i) {
		port_t *port;
		for(port = self->ports[PORT_INPUT][i]; port; port=port->next) {
			port->jack_buf = jack_port_get_buffer(port->jack_port, nframes);
			jack_midi_clear_buffer(port->jack_buf, nframes);
		}
	}

	while (1) {
		jack_ringbuffer_data_t vec[2];
		midi_event_t event;

		jack_ringbuffer_get_read_vector(self->midi_in, vec);
		if (jack_ringbuffer_vector_space(vec) < sizeof(event))
			break; // ENODATA
		jack_ringbuffer_vector_read(vec, &event, sizeof(event));
		if (jack_ringbuffer_vector_space(vec) < event.size)
			break; // ENODATA
		if (event.size > sizeof(self->jack_event_buf)) {
			jack_ringbuffer_read_advance(self->midi_in, sizeof(event) + event.size);
			break; // ENOSPC
		}

		jack_ringbuffer_vector_read(vec, self->jack_event_buf, event.size);
		jack_ringbuffer_read_advance(self->midi_in, sizeof(event) + event.size);

		if (event.time < jack_time) {
			event.time = jack_time;
			printf("time shift\n");
		}
#if 1
		if (jack_midi_event_write(event.port->jack_buf, event.time - jack_time, self->jack_event_buf, event.size, nframes))
			printf("ENOBUFS for %d bytes\n", (int)event.size);
#endif
	}
}

static void jack_output_midi(alsa_seqmidi_t *self, port_t *port, jack_nframes_t nframes)
{
	jack_nframes_t cur_frame_time = jack_last_frame_time(self->jack_client) + nframes;
	void *buf = jack_port_get_buffer(port->jack_port, nframes);
	jack_nframes_t ev_index;
	jack_nframes_t ev_count = jack_midi_get_event_count(buf, nframes);

	for (ev_index = 0; ev_index < ev_count; ++ev_index) {
		jack_midi_event_t jack_event;
		jack_ringbuffer_data_t vec[2];
		midi_event_t event;

		if (jack_midi_event_get(&jack_event, buf, ev_index, nframes))
			break; // ENODATA
		jack_ringbuffer_get_write_vector(self->midi_out, vec);
		if (jack_ringbuffer_vector_space(vec) < (sizeof(event) + jack_event.size))
			break; // ENOSPC

		event.port = port;
		event.time = jack_event.time + cur_frame_time;
		event.size = jack_event.size;
		jack_ringbuffer_vector_write(vec, &event, sizeof(event));
		jack_ringbuffer_vector_write(vec, jack_event.buffer, jack_event.size);
		jack_ringbuffer_write_advance(self->midi_out, sizeof(event) + event.size);
	}
}

static void jack_process_output(alsa_seqmidi_t *self, jack_nframes_t nframes)
{
	int i;
	for(i=0; i<PORT_HASH_SIZE; ++i) {
		port_t *port;
		for(port = self->ports[PORT_OUTPUT][i]; port; port=port->next)
			jack_output_midi(self, port, nframes);
	}
}

static int jack_process(jack_nframes_t nframes, void *arg)
{
	alsa_seqmidi_t *self = arg;
	// pass dead ports back to alsa thread
#if 1
	while (jack_ringbuffer_read_space(self->dead4alsa) >= sizeof(port_t*) && jack_ringbuffer_write_space(self->dead4jack) >= sizeof(port_t*)) {
		port_t *port;
		jack_ringbuffer_read(self->dead4alsa, (char*)&port, sizeof(port));
		jack_ringbuffer_write(self->dead4jack, (char*)&port, sizeof(port));
	}
#else
	jack_ringbuffer_data_t rd[2], wr[2];
	jack_ringbuffer_get_read_vector(dead4alsa, wr);
	jack_ringbuffer_get_write_vector(dead4jack, wr);
	size_t todo = jack_ringbuffer_vector_space(wr) < jack_ringbuffer_vector_space(rd)
		? jack_ringbuffer_vector_space(wr) : jack_ringbuffer_vector_space(rd);
	todo &= ~(sizeof(port_t*) - 1);
	if (todo)
		jack_ringbuffer_vector_write(wr, rd[0].buf, rd[0].len < todo ? rd[0].len : todo);
	if (todo > rd[0].len )
		jack_ringbuffer_vector_write(wr, rd[1].buf, todo - rd[0].len);
	jack_ringbuffer_read_advance(dead4alsa, todo);
	jack_ringbuffer_write_avance(dead4jack, todo);
#endif
	jack_process_input(self, nframes);
	jack_process_output(self, nframes);
	sem_post(&self->alsa_sem); // wakeup alsa thread
	return 0;
}

static int keep_running = 1;

void* alsa_process(void *arg)
{
	alsa_seqmidi_t *self = arg;
	while (keep_running) {
		alsa_input(self);
		alsa_output(self);
		snd_seq_drain_output(self->alsa_client);
		alsa_ports_cleanup(self);
		sem_wait(&self->alsa_sem);
	}		
	return 0;
}

static void jack_shutdown(void *arg)
{
	keep_running = 0;
}

void sigint_handler(int i)
{
	keep_running = 0;
}

int alsa_midi_start(alsa_seqmidi_t *self, const char *client_name)
{
	snd_seq_port_subscribe_t *sub;
	snd_seq_addr_t addr;

	if (snd_seq_open(&self->alsa_client, "hw", SND_SEQ_OPEN_DUPLEX, 0) < 0) {
		printf("failed to open alsa seq");
		return -1;
	}
	snd_seq_set_client_name(self->alsa_client, client_name);
	self->alsa_announce = snd_seq_create_simple_port(self->alsa_client, "update",
		SND_SEQ_PORT_CAP_READ|SND_SEQ_PORT_CAP_WRITE|SND_SEQ_PORT_CAP_SUBS_WRITE|SND_SEQ_PORT_CAP_NO_EXPORT,
		SND_SEQ_PORT_TYPE_APPLICATION);

  	self->alsa_input_que = snd_seq_alloc_queue(self->alsa_client);
  	snd_seq_start_queue(self->alsa_client, self->alsa_input_que, 0); 

  	self->alsa_output_que = snd_seq_alloc_queue(self->alsa_client);
  	snd_seq_start_queue(self->alsa_client, self->alsa_output_que, 0); 

	snd_seq_port_subscribe_alloca(&sub);
	addr.client = SND_SEQ_CLIENT_SYSTEM;
	addr.port = SND_SEQ_PORT_SYSTEM_ANNOUNCE;
	snd_seq_port_subscribe_set_sender(sub, &addr);
	addr.client = snd_seq_client_id(self->alsa_client);
	addr.port = self->alsa_announce;
	snd_seq_port_subscribe_set_dest(sub, &addr);
	snd_seq_port_subscribe_set_queue(sub, self->alsa_input_que);
	snd_seq_subscribe_port(self->alsa_client, sub);

	snd_seq_nonblock(self->alsa_client, 1);
	return 0;
}

int main()
{
	alsa_seqmidi_t *self = calloc(1, sizeof(alsa_seqmidi_t));
	if (alsa_midi_start(self, "jack_midi"))
		return 1;

	self->jack_client = jack_client_new("alsa_midi");
	if( !self->jack_client ) {
		printf("can't create jack client");
		goto fail1;
	}

	self->midi_in = jack_ringbuffer_create(64*1024);
	self->midi_out = jack_ringbuffer_create(64*1024);
	self->dead4alsa = jack_ringbuffer_create(64*sizeof(port_t*));
	self->dead4jack = jack_ringbuffer_create(64*sizeof(port_t*));
	snd_midi_event_new(1024, &self->alsa_decoder);
	snd_midi_event_init(self->alsa_decoder);
	sem_init(&self->alsa_sem, 0, 0);

	jack_on_shutdown(self->jack_client, jack_shutdown, 0);
	jack_set_process_callback(self->jack_client, jack_process, self);

	if (jack_activate(self->jack_client)) {
		printf("can't activate jack client");
		goto fail2;
	}

	signal(SIGINT, &sigint_handler);
	signal(SIGTERM, &sigint_handler);

	alsa_get_ports(self);
	alsa_process(self);

fail2:
	jack_client_close(self->jack_client);
fail1:
	snd_seq_close(self->alsa_client);
	return 0;
}
